==> peekabot-0.6.0.tar.bz2: peekabot-0.6.0/src/gui/Manipulator.cc <==
/**
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <iostream>
#include "../MessageHub.hh"
#include "../renderer/Renderer.hh"
#include "../IDTranslationTable.hh"
#include "../actions/PanCamera.hh"
#include "../actions/RotateCamera.hh"
#include "../actions/SetZoomDistance.hh"
#include "../actions/ViewportMove.hh"
#include "../actions/ViewportRotate.hh"
#include "Manipulator.hh"
#include "Viewport.hh"
#include "GUITypes.hh"
#include "UserInterfaceUtilities.hh"

using namespace peekabot;
using namespace gui;
using namespace boost::posix_time;
using namespace transformation_target;

Manipulator::Manipulator()
    : m_mode(NAVIGATE_MODE),
      m_system(CAMERA_COORDINATES),
      m_axes((Axis)(Y_AXIS |
                                           Z_AXIS)),
      m_pivot(AVERAGE_CENTER)
{

}

void Manipulator::set_mode(ManipulatorMode mode) throw()
{
    if(mode == m_mode)
    {
        rotate_system_option();
    }
    else
    {
        m_mode = mode;
        m_mode_signal(m_mode);

        set_system(CAMERA_COORDINATES);

        if( m_mode == ROTATE_MODE )
            set_pivot(AVERAGE_CENTER);
    }
}

void Manipulator::toggle_mode(ManipulatorMode mode) throw()
{
    if(mode == m_mode)
        set_mode(NAVIGATE_MODE);
    else
        set_mode(mode);
}

ManipulatorMode Manipulator::get_mode() const throw()
{
    return m_mode;
}

void Manipulator::set_system(CoordinateSystem system)
{
    m_system = system;         
    m_system_signal(m_system);
    if(m_mode == MOVE_MODE)
    {
        if(m_system == CAMERA_COORDINATES)
        {
            set_axes((Axis)(
                         Y_AXIS |
                         Z_AXIS));
        }
        else
            set_axes((Axis)(
                         X_AXIS |
                         Y_AXIS));
    }
    else if(m_mode == ROTATE_MODE)
    {
        if(m_system == CAMERA_COORDINATES)
            set_axes(X_AXIS);
        else
            set_axes(Z_AXIS);
    }
}

CoordinateSystem Manipulator::get_system()
    const throw()
{
    return m_system;
}

void Manipulator::set_pivot(RotationCenterpoint pivot)
{
    m_pivot = pivot;
    m_pivot_signal(m_pivot);
}

RotationCenterpoint Manipulator::get_pivot() const throw()
{
    return m_pivot;
}


void Manipulator::set_axes(Axis axes)
{
    // If we're in move mode with camera coordinates, clear the
    // x bit.
    if(m_mode == MOVE_MODE &&
       m_system == CAMERA_COORDINATES)
        axes = (Axis)(axes & (~X_AXIS));

    // Disallow zero axes.
    if(axes == 0)
        return;

    m_axes = axes;
    m_axis_signal(m_axes);
}

void Manipulator::toggle_axis(Axis axis)
{
    m_axes = (Axis)(m_axes ^ axis);
    m_axis_signal(m_axes);
}

Axis Manipulator::get_axes() const throw()
{
    return m_axes;
}


void Manipulator::viewport_click(Viewport *vp, int x, int y,
                                 int button, int state)
{
    if(button == MOUSE_LEFT_BUTTON)
        select(vp, x, y, button, state);
    else if(button == MOUSE_RIGHT_BUTTON && m_mode != NAVIGATE_MODE)
        set_mode(NAVIGATE_MODE);
    else if(button == MOUSE_RIGHT_BUTTON)
        TheUIUtilities::instance().clear_selection();
}

void Manipulator::viewport_drag(Viewport *vp, int dx, int dy,
                                int button, int state)
{
    switch(m_mode)
    {
        case NAVIGATE_MODE:
            navigate_mode(vp, dx, dy, button, state);
            break;
        case MOVE_MODE:
            move_mode(vp, dx, dy, button, state);
            break;
        case ROTATE_MODE:
            rotate_mode(vp, dx, dy, button, state);
            break;
        default:
            break;
    }    
}

void Manipulator::viewport_move(Viewport *vp, int dx, int dy, int state)
{

}

void Manipulator::viewport_scroll(Viewport *vp, int delta, int state)
{
    if(m_mode == NAVIGATE_MODE)
    {
        float rate = 1;
        if(state & KEY_SHIFT)
            rate /= 3;
        if(state & KEY_CTRL)
            rate *= 3;
        
        TheUIUtilities::instance().post_action(
            new SetZoomDistance(vp->get_camera(), delta / rate, true));
    }
}

int Manipulator::select(Viewport *vp, int x, int y, int button, int state)
{
    if(button == MOUSE_LEFT_BUTTON)
    {
        ObjectID click_object_id = renderer::TheRenderer::instance().pick(
            vp->w(), vp->h(), x, y, 
            TheIDTranslationTable::instance().get_primitive_id(vp->get_camera()), 
            vp->get_visible_layers(), 1);

        if(click_object_id == 0) // If it's 0, no object was picked
            return 1;

        /*
        TheMessageHub::instance().publish(
            INFO_MESSAGE, 
            (boost::format("Object %1% picked.")
             % m_click_object_id).str());
        */

        // Check if any keyboard buttons were held down.
        if(state & KEY_SHIFT)
        {
            TheUIUtilities::instance().select_object(
                click_object_id);
        }
        else if(state & KEY_CTRL)
        {
            TheUIUtilities::instance().toggle_select(
                click_object_id);
        }
        else
        {
            TheUIUtilities::instance().clear_selection();
            TheUIUtilities::instance().select_object(
                click_object_id);
        }
    }    
    
    return 1;
}

/* Free mode camera
 * 
 * During this mode of operation, mouse events control camera
 * movement only.
 *
 * Left-dragging pans the camera.
 * Right-dragging rotates the camera.
 * Middle button-dragging or scrolling the mouse wheel zooms the camera.
 */
int Manipulator::navigate_mode(Viewport *vp, int dx, int dy,
                               int button, int state)
{
    float rate = 1/1000.0;
    if(state & KEY_SHIFT)
        rate *= 5;
    if(state & KEY_CTRL)
        rate /= 5;

    switch(button)
    {
        case MOUSE_LEFT_BUTTON:
            // Left equals pan
            TheUIUtilities::instance().post_action(
                new PanCamera(vp->get_camera(), dx * rate, dy * rate));
            return 1;
        case MOUSE_MIDDLE_BUTTON:
            // Middle is zoom
            TheUIUtilities::instance().post_action(
                new SetZoomDistance(vp->get_camera(), dy * rate * 50, true));
            return 1;
        case MOUSE_RIGHT_BUTTON:
            TheUIUtilities::instance().post_action(
                new RotateCamera(vp->get_camera(), dx * rate / 5, dy * rate / 5));
            return 1;
        default:
            return 0;
    }
}

int Manipulator::move_mode(Viewport *vp, int dx, int dy, int button,
                             int state)
{    
    if(button == MOUSE_LEFT_BUTTON)
    {
        float rate = 1.0/50.0;
        if(state & KEY_SHIFT)
            rate *= 5;
        if(state & KEY_CTRL)
            rate /= 5;
        
        std::set<ObjectID> objects =
        TheUIUtilities::instance().get_selection();        
        for(std::set<ObjectID>::iterator it = objects.begin();
            it != objects.end(); ++it)
        {
            TheUIUtilities::instance().post_action(
                new ViewportMove(*it, dx*rate, dy*rate, m_axes,
                                 m_system, vp->get_camera()));
        }
        return 1;
    }
    else
        return 0;
}

int Manipulator::rotate_mode(Viewport *vp, int dx, int dy, int button,
                               int state)
{    
    if(button == MOUSE_LEFT_BUTTON)
    {
        float rate = M_PI / 1800 / 3;
        if(state & KEY_SHIFT)
            rate *= 5;
        if(state & KEY_CTRL)
            rate /= 5;
        
        std::set<ObjectID> objects =
            TheUIUtilities::instance().get_selection();        
        if(!objects.empty())
        {
            TheUIUtilities::instance().post_action(
                new ViewportRotate(objects, dx*rate, dy*rate, m_axes,
                                   m_system, m_pivot, vp->get_camera()));
        }
        return 1;
    }
    else
        return 0;
}


Manipulator::ModeSignal & Manipulator::get_mode_signal()
{
    return m_mode_signal;
}

Manipulator::SystemSignal & Manipulator::get_system_signal()
{
    return m_system_signal;
}

Manipulator::AxisSignal & Manipulator::get_axis_signal()
{
    return m_axis_signal;
}

Manipulator::PivotSignal & Manipulator::get_pivot_signal()
{
    return m_pivot_signal;
}

==> peekabot-0.7.2.tar.bz2: peekabot-0.7.2/src/gui/Manipulator.cc <==
/**
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <iostream>
#include "../MessageHub.hh"
#include "../RendererRelay.hh"
#include "../actions/PanCamera.hh"
#include "../actions/RotateCamera.hh"
#include "../actions/SetZoomDistance.hh"
#include "../actions/ViewportMove.hh"
#include "../actions/ViewportRotate.hh"
#include "Manipulator.hh"
#include "Viewport.hh"
#include "GUITypes.hh"
#include "UserInterfaceUtilities.hh"

using namespace peekabot;
using namespace gui;
using namespace boost::posix_time;
using namespace transformation_target;

Manipulator::Manipulator()
    : m_mode(NAVIGATE_MODE),
      m_system(CAMERA_COORDINATES),
      m_axes((Axis)(Y_AXIS |
                                           Z_AXIS)),
      m_pivot(AVERAGE_CENTER)
{

}

void Manipulator::set_mode(ManipulatorMode mode) throw()
{
    if(mode == m_mode)
    {
        rotate_system_option();
    }
    else
    {
        m_mode = mode;
        m_mode_signal(m_mode);

        set_system(CAMERA_COORDINATES);

        if( m_mode == ROTATE_MODE )
            set_pivot(AVERAGE_CENTER);
    }
}

void Manipulator::toggle_mode(ManipulatorMode mode) throw()
{
    if(mode == m_mode)
        set_mode(NAVIGATE_MODE);
    else
        set_mode(mode);
}

ManipulatorMode Manipulator::get_mode() const throw()
{
    return m_mode;
}

void Manipulator::set_system(CoordinateSystem system)
{
    m_system = system;         
    m_system_signal(m_system);
    if(m_mode == MOVE_MODE)
    {
        if(m_system == CAMERA_COORDINATES)
        {
            set_axes((Axis)(
                         Y_AXIS |
                         Z_AXIS));
        }
        else
            set_axes((Axis)(
                         X_AXIS |
                         Y_AXIS));
    }
    else if(m_mode == ROTATE_MODE)
    {
        if(m_system == CAMERA_COORDINATES)
            set_axes(X_AXIS);
        else
            set_axes(Z_AXIS);
    }
}

CoordinateSystem Manipulator::get_system()
    const throw()
{
    return m_system;
}

void Manipulator::set_pivot(RotationCenterpoint pivot)
{
    m_pivot = pivot;
    m_pivot_signal(m_pivot);
}

RotationCenterpoint Manipulator::get_pivot() const throw()
{
    return m_pivot;
}


void Manipulator::set_axes(Axis axes)
{
    // If we're in move mode with camera coordinates, clear the
    // x bit.
    if(m_mode == MOVE_MODE &&
       m_system == CAMERA_COORDINATES)
        axes = (Axis)(axes & (~X_AXIS));

    // Disallow zero axes.
    if(axes == 0)
        return;

    m_axes = axes;
    m_axis_signal(m_axes);
}

void Manipulator::toggle_axis(Axis axis)
{
    m_axes = (Axis)(m_axes ^ axis);
    m_axis_signal(m_axes);
}

Axis Manipulator::get_axes() const throw()
{
    return m_axes;
}


void Manipulator::viewport_click(Viewport *vp, int x, int y,
                                 int button, int state)
{
    if(button == MOUSE_LEFT_BUTTON)
        select(vp, x, y, button, state);
    else if(button == MOUSE_RIGHT_BUTTON && m_mode != NAVIGATE_MODE)
        set_mode(NAVIGATE_MODE);
    else if(button == MOUSE_RIGHT_BUTTON)
        TheUIUtilities::instance().clear_selection();
}

void Manipulator::viewport_drag(Viewport *vp, int dx, int dy,
                                int button, int state)
{
    switch(m_mode)
    {
        case NAVIGATE_MODE:
            navigate_mode(vp, dx, dy, button, state);
            break;
        case MOVE_MODE:
            move_mode(vp, dx, dy, button, state);
            break;
        case ROTATE_MODE:
            rotate_mode(vp, dx, dy, button, state);
            break;
        default:
            break;
    }    
}

void Manipulator::viewport_move(Viewport *vp, int dx, int dy, int state)
{

}

void Manipulator::viewport_scroll(Viewport *vp, int delta, int state)
{
    if(m_mode == NAVIGATE_MODE)
    {
        float rate = 1;
        if(state & KEY_SHIFT)
            rate /= 3;
        if(state & KEY_CTRL)
            rate *= 3;
        
        TheUIUtilities::instance().post_action(
            new SetZoomDistance(vp->get_camera(), delta / rate, true));
    }
}

int Manipulator::select(Viewport *vp, int x, int y, int button, int state)
{
    if(button == MOUSE_LEFT_BUTTON)
    {
        ObjectID click_object_id = TheRendererRelay::instance().pick(
            vp->w(), vp->h(), 
            vp->get_camera(), 
            vp->get_visible_layers(), 
            x, y, 1);

        if(click_object_id == 0) // If it's 0, no object was picked
            return 1;

        /*
        TheMessageHub::instance().publish(
            INFO_MESSAGE, 
            (boost::format("Object %1% picked.")
             % m_click_object_id).str());
        */

        // Check if any keyboard buttons were held down.
        if(state & KEY_SHIFT)
        {
            TheUIUtilities::instance().select_object(
                click_object_id);
        }
        else if(state & KEY_CTRL)
        {
            TheUIUtilities::instance().toggle_select(
                click_object_id);
        }
        else
        {
            TheUIUtilities::instance().clear_selection();
            TheUIUtilities::instance().select_object(
                click_object_id);
        }
    }    
    
    return 1;
}

/* Free mode camera
 * 
 * During this mode of operation, mouse events control camera
 * movement only.
 *
 * Left-dragging pans the camera.
 * Right-dragging rotates the camera.
 * Middle button-dragging or scrolling the mouse wheel zooms the camera.
 */
int Manipulator::navigate_mode(Viewport *vp, int dx, int dy,
                               int button, int state)
{
    float rate = 1/1000.0;
    if(state & KEY_SHIFT)
        rate *= 5;
    if(state & KEY_CTRL)
        rate /= 5;

    switch(button)
    {
        case MOUSE_LEFT_BUTTON:
            // Left equals pan
            TheUIUtilities::instance().post_action(
                new PanCamera(vp->get_camera(), dx * rate, dy * rate));
            return 1;
        case MOUSE_MIDDLE_BUTTON:
            // Middle is zoom
            TheUIUtilities::instance().post_action(
                new SetZoomDistance(vp->get_camera(), dy * rate * 50, true));
            return 1;
        case MOUSE_RIGHT_BUTTON:
            TheUIUtilities::instance().post_action(
                new RotateCamera(vp->get_camera(), dx * rate / 5, dy * rate / 5));
            return 1;
        default:
            return 0;
    }
}

int Manipulator::move_mode(Viewport *vp, int dx, int dy, int button,
                             int state)
{    
    if(button == MOUSE_LEFT_BUTTON)
    {
        float rate = 1.0/50.0;
        if(state & KEY_SHIFT)
            rate *= 5;
        if(state & KEY_CTRL)
            rate /= 5;
        
        std::set<ObjectID> objects =
        TheUIUtilities::instance().get_selection();        
        for(std::set<ObjectID>::iterator it = objects.begin();
            it != objects.end(); ++it)
        {
            TheUIUtilities::instance().post_action(
                new ViewportMove(*it, dx*rate, dy*rate, m_axes,
                                 m_system, vp->get_camera()));
        }
        return 1;
    }
    else
        return 0;
}

int Manipulator::rotate_mode(Viewport *vp, int dx, int dy, int button,
                               int state)
{    
    if(button == MOUSE_LEFT_BUTTON)
    {
        float rate = M_PI / 1800 / 3;
        if(state & KEY_SHIFT)
            rate *= 5;
        if(state & KEY_CTRL)
            rate /= 5;
        
        std::set<ObjectID> objects =
            TheUIUtilities::instance().get_selection();        
        if(!objects.empty())
        {
            TheUIUtilities::instance().post_action(
                new ViewportRotate(objects, dx*rate, dy*rate, m_axes,
                                   m_system, m_pivot, vp->get_camera()));
        }
        return 1;
    }
    else
        return 0;
}


Manipulator::ModeSignal & Manipulator::get_mode_signal()
{
    return m_mode_signal;
}

Manipulator::SystemSignal & Manipulator::get_system_signal()
{
    return m_system_signal;
}

Manipulator::AxisSignal & Manipulator::get_axis_signal()
{
    return m_axis_signal;
}

Manipulator::PivotSignal & Manipulator::get_pivot_signal()
{
    return m_pivot_signal;
}
